use std::fmt::{Display, Formatter};
use std::str;

#[derive(Clone, Debug, Eq, Hash, PartialEq)]
pub struct Error(pub(crate) ErrorKind);

#[derive(Clone, Debug, Eq, Hash, PartialEq)]
pub(crate) enum ErrorKind {
    ByteLength { len: usize },
    Char { character: char, index: usize },
    GroupCount { count: usize },
    GroupLength { group: usize, len: usize, index: usize },
    InvalidUtf8,
}

#[derive(Clone, Debug, Eq, Hash, PartialEq)]
pub struct InvalidUuid<'a>(pub(crate) &'a [u8]);

impl<'a> InvalidUuid<'a> {
    pub fn into_err(self) -> Error {
        // Check if the string is valid UTF-8
        let uuid_str = match str::from_utf8(self.0) {
            Ok(s) => s,
            Err(_) => return Error(ErrorKind::InvalidUtf8)
        };

        let mut hyphen_count = 0;
        let mut group_bounds = [0; 4];

        let uuid_str = unsafe { str::from_utf8_unchecked(uuid_str.as_bytes()) };

        for (index, character) in uuid_str.char_indices() {
            let byte = character as u8;
            if character as u32 - byte as u32 > 0 {
                // The char is more than one byte
                return Error(ErrorKind::Char {
                    character,
                    index: index + 1,
                });
            } else if byte == b'-' {
                // Search for the groups bounds
                if hyphen_count < 4 {
                    group_bounds[hyphen_count] = index;
                }
                hyphen_count += 1;
            } else if !matches!(byte, b'0'..=b'9' | b'a'..=b'f' | b'A'..=b'F') {
                // Non hex char
                return Error(ErrorKind::Char {
                    character,
                    index: index + 1,
                });
            }
        }

        if hyphen_count != 4 {
            // There is more or less than 5 groups
            Error(ErrorKind::GroupCount {
                count: hyphen_count + 1
            })
        } else {
            // One of the groups as an invalid length
            const BLOCK_STARTS: [usize; 5] = [0, 9, 14, 19, 24];
            for i in 0..4 {
                if group_bounds[i] != BLOCK_STARTS[i + 1] - 1 {
                    return Error(ErrorKind::GroupLength {
                        group: i,
                        len: group_bounds[i] - BLOCK_STARTS[i],
                        index: BLOCK_STARTS[i] + 1,
                    });
                }
            }

            // The last group is too long
            Error(ErrorKind::GroupLength {
                group: 4,
                len: uuid_str.len() - BLOCK_STARTS[4],
                index: BLOCK_STARTS[4] + 1,
            })
        }
    }
}

impl Display for Error {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        match self.0 {
            ErrorKind::ByteLength { len } => {
                write!(f, "invalid length: expected 16 bytes, found {}", len)
            }
            ErrorKind::Char { character, index, .. } => {
                write!(f, "invalid character: expected a sequence of [0-9a-fA-F-], found `{}` at {}", character, index)
            }
            ErrorKind::GroupCount { count } => {
                write!(f, "invalid group count: expected 5, found {}", count)
            }
            ErrorKind::GroupLength { group, len, .. } => {
                let expected = [8, 4, 4, 4, 12][group];
                write!(f, "invalid group length in group {}: epected {}, found {}", group, expected, len)
            }
            ErrorKind::InvalidUtf8 => write!(f, "non UTF-8 input"),
        }
    }
}
