use std::str::FromStr;
use crate::error::{Error, InvalidUuid};
use crate::Uuid;

impl FromStr for Uuid {
    type Err = Error;

    fn from_str(value: &str) -> Result<Self, Self::Err> {
        Uuid::parse_str(value)
    }
}

impl TryFrom<&'_ str> for Uuid {
    type Error = Error;

    fn try_from(value: &'_ str) -> Result<Self, Self::Error> {
        Uuid::parse_str(value)
    }
}

impl Uuid {
    pub fn parse_str(value: &str) -> Result<Self, Error> {
        try_parse(value.as_bytes())
            .map(Uuid::from_bytes)
            .map_err(InvalidUuid::into_err)
    }
}

const fn try_parse(input: &[u8]) -> Result<[u8; 16], InvalidUuid> {
    let result = match (input.len(), input) {
        (36, s) => parse(s),
        _ => Err(())
    };

    match result {
        Ok(b) => Ok(b),
        Err(()) => Err(InvalidUuid(input))
    }
}

const fn parse(s: &[u8]) -> Result<[u8; 16], ()> {
    if s.len() != 36 {
        return Err(());
    }

    // Checks that the string looks like a valid UUID
    match [s[8], s[13], s[18], s[23]] {
        [b'-', b'-', b'-', b'-'] => {}
        _ => return Err(())
    }

    let positions: [u8; 8] = [0, 4, 9, 14, 19, 24, 28, 32];
    let mut buf: [u8; 16] = [0; 16];
    let mut j = 0;

    while j < 8 {
        let i = positions[j];

        let h1 = HEX_TABLE[s[i as usize] as usize];
        let h2 = HEX_TABLE[s[(i + 1) as usize] as usize];
        let h3 = HEX_TABLE[s[(i + 2) as usize] as usize];
        let h4 = HEX_TABLE[s[(i + 3) as usize] as usize];

        if h1 | h2 | h3 | h4 == 0xff {
            return Err(());
        }

        buf[j * 2] = SHL4_TABLE[h1 as usize] | h2;
        buf[j * 2 + 1] = SHL4_TABLE[h3 as usize] | h4;
        j += 1;
    }

    Ok(buf)
}

const HEX_TABLE: &[u8; 256] = &{
    let mut buf = [0u8; 256];
    let mut i = 0u8;

    loop {
        buf[i as usize] = match i {
            b'0'..=b'9' => i - b'0',
            b'a'..=b'f' => i - b'a' + 10,
            b'A'..=b'F' => i - b'A' + 10,
            _ => 0xff
        };

        if i == 255 {
            break buf;
        }

        i += 1
    }
};

// Shift left 4 bits lookup table
const SHL4_TABLE: &[u8; 256] = &{
    let mut buf = [0u8; 256];
    let mut i = 0u8;

    loop {
        buf[i as usize] = i.wrapping_shl(4);

        if i == 255 {
            break buf;
        }

        i += 1;
    }
};

#[cfg(test)]
mod tests {
    use crate::Uuid;

    #[test]
    fn test_parse_uuid_v4() {
        let uuid = Uuid::parse_str("67e55044-10b1-426f-9247-bb680e5fe0c8").unwrap();
        print!("var: {:?}, ver: {:?}, uuid: {}", uuid.get_variant(), uuid.get_version(), uuid.to_string());
    }
}
